# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn

from ..builder import DISTILLERS, build_loss
from .base import BaseDistiller


@DISTILLERS.register_module()
class SelfDistiller(BaseDistiller):
    """Transfer knowledge inside a single model.

    Args:
        components (dict): The details of the distillation. It usually includes
            the module names of the teacher and the student, and the losses
            used in the distillation.
    """

    def __init__(self, components, **kwargs):
        super().__init__(**kwargs)
        self.components = components
        self.losses = nn.ModuleDict()

        self.student_outputs = dict()
        self.teacher_outputs = dict()

        for component in self.components:
            student_module_name = component['student_module']
            teacher_module_name = component['teacher_module']
            self.student_outputs[student_module_name] = list()
            self.teacher_outputs[teacher_module_name] = list()

            for loss in component.losses:
                loss_cfg = loss.copy()
                loss_name = loss_cfg.pop('name')
                self.losses[loss_name] = build_loss(loss_cfg)

    def prepare_from_student(self, student):
        """Registers a global forward hook for each teacher module and student
        module to be used in the distillation.

        Args:
            student (:obj:`torch.nn.Module`): The student model to be used
                in the distillation.
        """
        self.module2name = {}
        for name, module in student.model.named_modules():
            self.module2name[module] = name
        self.name_modules = dict(student.model.named_modules())

        for component in self.components:
            student_module_name = component['student_module']
            teacher_module_name = component['teacher_module']

            student_module = self.name_modules[student_module_name]
            teacher_module = self.name_modules[teacher_module_name]

            student_module.register_forward_hook(
                self.student_forward_output_hook)
            teacher_module.register_forward_hook(
                self.teacher_forward_output_hook)

    def teacher_forward_output_hook(self, module, inputs, outputs):
        """Save the output.

        Args:
            module (:obj:`torch.nn.Module`): the module of register hook
            inputs (tuple): input of module
            outputs (tuple): out of module
        """
        if self.training and getattr(self, 'is_teacher', None):
            self.teacher_outputs[self.module2name[module]].append(outputs)

    def student_forward_output_hook(self, module, inputs, outputs):
        """Save the output.

        Args:
            module (:obj:`torch.nn.Module`): the module of register hook
            inputs (tuple): input of module
            outputs (tuple): out of module
        """
        if self.training and not getattr(self, 'is_teacher', None):
            self.student_outputs[self.module2name[module]].append(outputs)

    def reset_outputs(self, outputs):
        """Reset the teacher's outputs or student's outputs."""
        for key in outputs.keys():
            outputs[key] = list()

    def exec_teacher_forward(self, teacher, data):
        """Forward computation of the teacher.

        Args:
            teacher (:obj:`torch.nn.Module`): The teacher model to be used
                in the distillation.
            data (dict): The output of dataloader.
        """
        self.reset_outputs(self.teacher_outputs)
        self.is_teacher = True
        output = teacher(**data)
        self.is_teacher = False

        return output

    def exec_student_forward(self, student, data):
        """Forward computation of the student.

        Args:
            student (:obj:`torch.nn.Module`): The student model to be used
                in the distillation.
            data (dict): The output of dataloader.
        """
        assert not self.is_teacher
        self.reset_outputs(self.student_outputs)
        output = student(**data)

        return output

    def compute_distill_loss(self, data):
        """Compute the distillation loss."""

        losses = dict()

        for i, component in enumerate(self.components):
            student_module_name = component['student_module']
            student_outputs = self.student_outputs[student_module_name]

            teacher_module_name = component['teacher_module']
            teacher_outputs = self.teacher_outputs[teacher_module_name]

            for out_idx, (s_out, t_out) in enumerate(
                    zip(student_outputs, teacher_outputs)):

                for loss in component.losses:
                    loss_module = self.losses[loss.name]
                    loss_name = f'{loss.name}.{out_idx}'

                    loss_module.current_data = data
                    losses[loss_name] = loss_module(s_out, t_out)
                    loss_module.current_data = None

        return losses



# Copyright (c) OpenMMLab. All rights reserved.
from .darts_backbone import DartsBackbone
from .searchable_mobilenet import SearchableMobileNet
from .searchable_shufflenet_v2 import SearchableShuffleNetV2

__all__ = ['DartsBackbone', 'SearchableShuffleNetV2', 'SearchableMobileNet']



# Copyright (c) OpenMMLab. All rights reserved.
from .align_method_kd import AlignMethodDistill
from .autoslim import AutoSlim
from .darts import Darts
from .detnas import DetNAS
from .general_distill import GeneralDistill
from .spos import SPOS

__all__ = [
    'AutoSlim', 'AlignMethodDistill', 'Darts', 'SPOS', 'DetNAS',
    'GeneralDistill'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .components import *  # noqa: F401,F403
from .mmcls import MMClsArchitecture
from .mmdet import MMDetArchitecture
from .mmseg import MMSegArchitecture
from .utils import *  # noqa: F401,F403

__all__ = ['MMClsArchitecture', 'MMDetArchitecture', 'MMSegArchitecture']



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Union

import mmcv
from mmcv.runner import load_checkpoint
from torch import nn

from mmrazor.models import build_algorithm


def init_mmseg_model(config: Union[str, mmcv.Config],
                     checkpoint: Optional[str] = None,
                     device: str = 'cuda:0') -> nn.Module:
    """Initialize a mmseg model from config file.

    Args:
        config (str or :obj:`mmcv.Config`): Config file path or the config
            object.
        checkpoint (str, optional): Checkpoint path. If left as None, the model
            will not load any weights.
        device (str, optional) CPU/CUDA device option. Default 'cuda:0'.
            Use 'cpu' for loading model on CPU.
    Returns:
        nn.Module: The constructed segmentor.
    """
    if isinstance(config, str):
        config = mmcv.Config.fromfile(config)
    elif not isinstance(config, mmcv.Config):
        raise TypeError('config must be a filename or Config object, '
                        'but got {}'.format(type(config)))

    model_cfg = config.algorithm.architecture.model
    model_cfg.pretrained = None
    model_cfg.train_cfg = None
    algorithm = build_algorithm(config.algorithm)
    model = algorithm.architecture.model

    if checkpoint is not None:
        checkpoint = load_checkpoint(model, checkpoint, map_location='cpu')
        model.CLASSES = checkpoint['meta']['CLASSES']
        model.PALETTE = checkpoint['meta']['PALETTE']
    model.cfg = config  # save the config in the model for convenience
    model.to(device)
    model.eval()

    return model



# Copyright (c) OpenMMLab. All rights reserved.
import copy

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..builder import MUTABLES
from .mutable_module import MutableModule


class MutableEdge(MutableModule):
    """Mutable Edge. In some NAS algorithms (Darts, AutoDeeplab, etc.), the
    connections between modules are searchable, such as the connections between
    a node and its previous nodes in Darts. ``MutableEdge`` has N modules to
    process N inputs respectively.

    Args:
        choices (torch.nn.ModuleDict): Unlike ``MutableOP``, there are already
            created modules in choices.
    """

    def __init__(self, choices, **kwargs):
        super(MutableEdge, self).__init__(**kwargs)
        assert isinstance(choices, nn.ModuleDict), \
            f'In {self.space_id}, the choices must be torch.nn.ModuleDict.'
        self.choices = choices
        self.choice_mask = self.build_choice_mask()
        # Once execute ``export``, the unchosen modules will be removed.
        # Record full choice names, which will be used in forward.
        self.full_choice_names = copy.deepcopy(self.choice_names)

    def build_choices(self, cfg):
        """MutableEdge's choices is already built."""
        pass


@MUTABLES.register_module()
class DifferentiableEdge(MutableEdge):
    """Differentiable Edge.

    Search the best module from choices by learnable parameters.

    Args:
        with_arch_param (bool): whether build learable architecture parameters.
    """

    def __init__(self, with_arch_param, **kwargs):
        super(DifferentiableEdge, self).__init__(**kwargs)
        self.with_arch_param = with_arch_param

    def build_arch_param(self):
        """build learnable architecture parameters."""
        if self.with_arch_param:
            return nn.Parameter(torch.randn(self.num_choices) * 1e-3)
        else:
            return None

    def compute_arch_probs(self, arch_param):
        """compute chosen probs according architecture parameters."""
        return F.softmax(arch_param, -1)

    def forward(self, prev_inputs, arch_param=None):
        """forward function.

        In some algorithms, there are several ``MutableModule`` share the same
        architecture parameters. So the architecture parameters are passed
        in as args.

        Args:
            prev_inputs (list[torch.Tensor]): each choice's inputs.
            arch_param (torch.nn.Parameter): architecture parameters.
        """

        assert len(self.full_choice_names) == len(prev_inputs), \
            f'In {self.space_id}, the length of full choice names must be \
                same as the length of previous inputs.'

        if self.with_arch_param:
            assert arch_param is not None, \
                f'In {self.space_id}, the arch_param can not be None when the \
                    with_arch_param=True.'

            # 1. compute choices' probs.
            probs = self.compute_arch_probs(arch_param)

            # 2. process each input
            outputs = list()
            for prob, module, input in zip(probs, self.choice_modules,
                                           prev_inputs):
                if prob > 0:
                    outputs.append(prob * module(input))

        else:
            outputs = list()
            for name, input in zip(self.full_choice_names, prev_inputs):
                if name not in self.choice_names:
                    continue
                module = self.choices[name]
                outputs.append(module(input))

            assert len(outputs) > 0

        return sum(outputs)


@MUTABLES.register_module()
class GumbelEdge(DifferentiableEdge):
    """Gumbel Edge.

    Search the best module from choices by gumbel trick.
    """

    def __init__(self, **kwargs):
        super(GumbelEdge, self).__init__(**kwargs)

    def set_temperature(self, temperature):
        """Modify the temperature."""
        self.temperature = temperature

    def compute_arch_probs(self, arch_param):
        """compute chosen probs by gumbel trick."""
        probs = F.gumbel_softmax(
            arch_param, tau=self.tau, hard=self.hard, dim=-1)
        return probs



# Copyright (c) OpenMMLab. All rights reserved.

import torch
from torch.nn import functional as F

from mmrazor.models.builder import MUTATORS
from .differentiable_mutator import DifferentiableMutator


@MUTATORS.register_module()
class DartsMutator(DifferentiableMutator):

    def __init__(self, ignore_choices=('zero', ), **kwargs):
        super().__init__(**kwargs)
        self.ignore_choices = ignore_choices

    def search_subnet(self):

        subnet_dict = dict()
        for space_id, sub_space in self.search_spaces.items():
            if space_id in self.arch_params:
                space_arch_param = self.arch_params[space_id]
                arch_probs = F.softmax(space_arch_param, dim=-1)
                choice_names = sub_space['choice_names']
                keep_idx = [
                    i for i, name in enumerate(choice_names)
                    if name not in self.ignore_choices
                ]
                best_choice_prob, best_choice_idx = torch.max(
                    arch_probs[keep_idx], 0)
                best_choice_idx = keep_idx[best_choice_idx.item()]
                best_choice_name = choice_names[best_choice_idx]

                subnet_dict[space_id] = dict(
                    chosen=[best_choice_name],
                    chosen_probs=[best_choice_prob.item()])

        def sort_key(x):
            return subnet_dict[x]['chosen_probs'][0]

        for space_id, sub_space in self.search_spaces.items():
            if space_id not in self.arch_params:
                num_chosen = sub_space['num_chosen']
                choice_names = sub_space['choice_names']
                sorted_edges = list(
                    sorted(choice_names, key=sort_key, reverse=True))
                chosen = sorted_edges[:num_chosen]
                subnet_dict[space_id] = dict(chosen=chosen)

                for not_chosen in sorted_edges[num_chosen:]:
                    subnet_dict.pop(not_chosen)

        return subnet_dict



# Copyright (c) Open-MMLab. All rights reserved.
import os.path as osp
import platform
import shutil
import time
from functools import partial

import mmcv
import yaml
from mmcv.runner import HOOKS, RUNNERS, EpochBasedRunner

from ..utils import set_lr


class EpochMultiLoader:
    """Multi loaders based on epoch."""

    def __init__(self, dataloaders):
        self._dataloaders = dataloaders
        self.iter_loaders = [iter(loader) for loader in self._dataloaders]

    @property
    def num_loaders(self):
        """The number of dataloaders."""
        return len(self._dataloaders)

    def __iter__(self):
        """Return self when executing __iter__."""
        return self

    def __next__(self):
        """Get next iter's data."""
        data = tuple([next(loader) for loader in self.iter_loaders])

        return data

    def __len__(self):
        """Get the length of loader."""
        return min([len(loader) for loader in self._dataloaders])


@RUNNERS.register_module()
class MultiLoaderEpochBasedRunner(EpochBasedRunner):
    """Multi Dataloaders EpochBasedRunner.

    There are three differences from EpochBaseRunner： 1）Support load data from
    multi dataloaders. 2) Support freeze some optimizer's lr update when runner
    has multi optimizers. 3) Add ``search_subnet`` api.
    """

    def train(self, data_loader, **kwargs):
        """Rewrite the ``train`` of ``EpochBasedRunner``."""
        self.model.train()
        self.mode = 'train'

        # the only difference from EpochBasedRunner's ``train``
        if isinstance(data_loader, list):
            self.data_loader = EpochMultiLoader(data_loader)
        else:
            self.data_loader = data_loader

        self._max_iters = self._max_epochs * len(self.data_loader)
        self.call_hook('before_train_epoch')
        time.sleep(2)  # Prevent possible deadlock during epoch transition
        for i, data_batch in enumerate(self.data_loader):
            self._inner_iter = i
            self.call_hook('before_train_iter')
            self.run_iter(data_batch, train_mode=True, **kwargs)
            self.call_hook('after_train_iter')
            self._iter += 1

        self.call_hook('after_train_epoch')

        self._epoch += 1

    def register_lr_hook(self, lr_config):
        """Resister a hook for setting learning rate.

        Args:
            lr_config (dict): Config for setting learning rate.
        """
        if lr_config is None:
            return
        elif isinstance(lr_config, dict):
            assert 'policy' in lr_config
            policy_type = lr_config.pop('policy')
            if 'freeze_optimizers' in lr_config:
                freeze_optimizers = lr_config.pop('freeze_optimizers')
            else:
                freeze_optimizers = []
            # If the type of policy is all in lower case, e.g., 'cyclic',
            # then its first letter will be capitalized, e.g., to be 'Cyclic'.
            # This is for the convenient usage of Lr updater.
            # Since this is not applicable for ``CosineAnnealingLrUpdater``,
            # the string will not be changed if it contains capital letters.
            if policy_type == policy_type.lower():
                policy_type = policy_type.title()
            hook_type = policy_type + 'LrUpdaterHook'
            lr_config['type'] = hook_type
            hook = mmcv.build_from_cfg(lr_config, HOOKS)

            # modify the hook's ``_set_lr``
            # the only difference from BasedRunner's ``register_lr_hook``
            hook._set_lr = partial(set_lr, freeze_optimizers=freeze_optimizers)
        else:
            hook = lr_config

        self.register_hook(hook, priority=10)

    def search_subnet(self,
                      out_dir,
                      filename_tmpl='epoch_{}.yaml',
                      create_symlink=True):
        """Search the best subnet.

        Args:
            out_dir (str): The directory that subnets are saved.
            filename_tmpl (str, optional): The subnet filename template,
                which contains a placeholder for the epoch number.
                Defaults to 'epoch_{}.yaml'.
            create_symlink (bool, optional): Whether to create a symlink
                "latest.yaml" to point to the latest subnet.
                Defaults to True.
        """

        filename = filename_tmpl.format(self.epoch + 1)
        filepath = osp.join(out_dir, filename)

        algorithm = self.model.module if hasattr(self.model,
                                                 'module') else self.model
        mutator = algorithm.mutator.module if hasattr(
            algorithm.mutator, 'module') else algorithm.mutator
        subnet = mutator.search_subnet()
        with open(filepath, 'w') as f:
            yaml.dump(subnet, f)
        # in some environments, ``os.symlink`` is not supported, you may need
        # to set ``create_symlink`` to False
        if create_symlink:
            dst_file = osp.join(out_dir, 'latest.yaml')
            if platform.system() != 'Windows':
                mmcv.symlink(filename, dst_file)
            else:
                shutil.copy(filepath, dst_file)



# Copyright (c) OpenMMLab. All rights reserved.
try:
    import mmdet
except (ImportError, ModuleNotFoundError):
    mmdet = None

if mmdet is not None:
    from .inference import init_mmdet_model
    from .train import set_random_seed, train_mmdet_model

    __all__ = ['train_mmdet_model', 'init_mmdet_model', 'set_random_seed']



# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.utils import Registry, build_from_cfg

SEARCHERS = Registry('search')


def build_searcher(cfg, default_args=None):
    """Build searcher."""
    return build_from_cfg(cfg, SEARCHERS, default_args=default_args)



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.utils.checkpoint as cp
from mmcls.models.utils import channel_shuffle
from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule

from ..builder import OPS
from .base import BaseOP


@OPS.register_module()
class ShuffleBlock(BaseOP):
    """InvertedResidual block for Searchable ShuffleNetV2 backbone.

    Args:
        kernel_size (int): Size of the convolving kernel.
        stride (int): Stride of the convolution layer. Default: 1
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='ReLU').
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.

    Returns:
        Tensor: The output tensor.
    """

    def __init__(self,
                 kernel_size,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 with_cp=False,
                 **kwargs):

        super(ShuffleBlock, self).__init__(**kwargs)

        assert kernel_size in [3, 5, 7]
        self.kernel_size = kernel_size
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.with_cp = with_cp

        branch_features = self.out_channels // 2
        if self.stride == 1:
            assert self.in_channels == branch_features * 2, (
                f'in_channels ({self.in_channels}) should equal to '
                f'branch_features * 2 ({branch_features * 2}) '
                'when stride is 1')

        if self.in_channels != branch_features * 2:
            assert self.stride != 1, (
                f'stride ({self.stride}) should not equal 1 when '
                f'in_channels != branch_features * 2')

        if self.stride > 1:
            self.branch1 = nn.Sequential(
                ConvModule(
                    self.in_channels,
                    self.in_channels,
                    kernel_size=self.kernel_size,
                    stride=self.stride,
                    padding=self.kernel_size // 2,
                    groups=self.in_channels,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=None),
                ConvModule(
                    self.in_channels,
                    branch_features,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg),
            )

        self.branch2 = nn.Sequential(
            ConvModule(
                self.in_channels if (self.stride > 1) else branch_features,
                branch_features,
                kernel_size=1,
                stride=1,
                padding=0,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg),
            ConvModule(
                branch_features,
                branch_features,
                kernel_size=self.kernel_size,
                stride=self.stride,
                padding=self.kernel_size // 2,
                groups=branch_features,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=None),
            ConvModule(
                branch_features,
                branch_features,
                kernel_size=1,
                stride=1,
                padding=0,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg))

    def forward(self, x):

        def _inner_forward(x):
            if self.stride > 1:
                out = torch.cat((self.branch1(x), self.branch2(x)), dim=1)
            else:
                x1, x2 = x.chunk(2, dim=1)
                out = torch.cat((x1, self.branch2(x2)), dim=1)

            out = channel_shuffle(out, 2)

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        return out


@OPS.register_module()
class ShuffleXception(BaseOP):
    """Xception block for ShuffleNetV2 backbone.

    Args:
        conv_cfg (dict, optional): Config dict for convolution layer.
            Defaults to None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Defaults to dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Defaults to dict(type='ReLU').
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Defaults to False.

    Returns:
        Tensor: The output tensor.
    """

    def __init__(self,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 with_cp=False,
                 **kwargs):
        super(ShuffleXception, self).__init__(**kwargs)
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.with_cp = with_cp
        self.mid_channels = self.out_channels // 2

        branch_features = self.out_channels // 2
        if self.stride == 1:
            assert self.in_channels == branch_features * 2, (
                f'in_channels ({self.in_channels}) should equal to '
                f'branch_features * 2 ({branch_features * 2}) '
                'when stride is 1')

        if self.in_channels != branch_features * 2:
            assert self.stride != 1, (
                f'stride ({self.stride}) should not equal 1 when '
                f'in_channels != branch_features * 2')

        if self.stride > 1:
            self.branch1 = nn.Sequential(
                ConvModule(
                    self.in_channels,
                    self.in_channels,
                    kernel_size=3,
                    stride=self.stride,
                    padding=1,
                    groups=self.in_channels,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=None),
                ConvModule(
                    self.in_channels,
                    branch_features,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg),
            )

        self.branch2 = []

        self.branch2.append(
            DepthwiseSeparableConvModule(
                self.in_channels if (self.stride > 1) else branch_features,
                self.mid_channels,
                kernel_size=3,
                stride=self.stride,
                padding=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                dw_act_cfg=None,
                act_cfg=self.act_cfg), )
        self.branch2.append(
            DepthwiseSeparableConvModule(
                self.mid_channels,
                self.mid_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                dw_act_cfg=None,
                act_cfg=self.act_cfg))
        self.branch2.append(
            DepthwiseSeparableConvModule(
                self.mid_channels,
                branch_features,
                kernel_size=3,
                stride=1,
                padding=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                dw_act_cfg=None,
                act_cfg=self.act_cfg))
        self.branch2 = nn.Sequential(*self.branch2)

    def forward(self, x):

        def _inner_forward(x):
            if self.stride > 1:
                out = torch.cat((self.branch1(x), self.branch2(x)), dim=1)
            else:
                x1, x2 = x.chunk(2, dim=1)
                out = torch.cat((x1, self.branch2(x2)), dim=1)

            out = channel_shuffle(out, 2)

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        return out



# Copyright (c) OpenMMLab. All rights reserved.
try:
    import mmdet
except (ImportError, ModuleNotFoundError):
    mmdet = None

try:
    import mmseg
except (ImportError, ModuleNotFoundError):
    mmseg = None

from mmcls.models import *  # noqa: F401,F403

from .algorithms import *  # noqa: F401,F403
from .architectures import *  # noqa: F401,F403
from .builder import (ALGORITHMS, ARCHITECTURES, DISTILLERS, LOSSES, MUTABLES,
                      MUTATORS, OPS, build_algorithm, build_architecture,
                      build_distiller, build_loss, build_mutable,
                      build_mutator, build_op)
from .distillers import *  # noqa: F401,F403
from .losses import *  # noqa: F401,F403
from .mutables import *  # noqa: F401,F403
from .mutators import *  # noqa: F401,F403
from .ops import *  # noqa: F401,F403
from .pruners import *  # noqa: F401,F403

if mmdet is not None:
    from mmdet.models import *  # noqa: F401,F403
if mmseg is not None:
    from mmseg.models import *  # noqa: F401,F403

__all__ = [
    'ALGORITHMS', 'MUTABLES', 'ARCHITECTURES', 'DISTILLERS', 'MUTATORS',
    'LOSSES', 'OPS', 'build_architecture', 'build_mutable', 'build_op',
    'build_mutator', 'build_algorithm', 'build_distiller', 'build_loss'
]



# Copyright (c) OpenMMLab. All rights reserved.
import copy

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..builder import MUTABLES, build_op
from .mutable_module import MutableModule


class MutableOP(MutableModule):
    """An important type of ``MUTABLES``, inherits from ``MutableModule``.

    Args:
        choices (dict): The configs for the choices, the chosen ``OPS`` used to
            combine ``MUTABLES``.
        choice_args (dict): The args used to set chosen ``OPS``.
    """

    def __init__(self, choices, choice_args, **kwargs):
        super(MutableOP, self).__init__(**kwargs)
        self.choices = self.build_choices(choices, choice_args)
        self.choice_mask = self.build_choice_mask()
        self.full_choice_names = copy.deepcopy(self.choice_names)

    def build_choices(self, cfgs, choice_args):
        """Build all chosen ``OPS`` used to combine ``MUTABLES``, and the
        choices will be sampled.

        Args:
            cfgs (dict): The configs for the choices.
            choice_args (dict): The args used to set chosen ``OPS``.

        Returns:
            torch.nn.ModuleDict: Consists of chosen ``OPS`` in the arg `cfgs`.
        """
        choices = nn.ModuleDict()
        for name, cfg in cfgs.items():
            cfg.update(choice_args)
            op_module = build_op(cfg)
            choices.add_module(name, op_module)
        return choices


@MUTABLES.register_module()
class OneShotOP(MutableOP):
    """A type of ``MUTABLES`` for the one-shot NAS."""

    def __init__(self, **kwargs):
        super(OneShotOP, self).__init__(**kwargs)
        assert self.num_chosen == 1

    def forward(self, x):
        """Forward computation for chosen ``OPS``, in one-shot NAS, the number
        of chosen ``OPS`` can only be one.

        Args:
            x (tensor | tuple[tensor]): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.

        Returns:
            torch.Tensor: The result of forward.
        """
        outputs = list()
        for name, chosen_bool in zip(self.full_choice_names, self.choice_mask):
            if name not in self.choice_names:
                continue
            if not chosen_bool:
                continue
            module = self.choices[name]
            outputs.append(module(x))

        assert len(outputs) > 0

        return sum(outputs)


@MUTABLES.register_module()
class DifferentiableOP(MutableOP):
    """Differentiable OP.

    Search the best module from choices by learnable parameters.

    Args:
        with_arch_param (bool): whether build learable architecture parameters.
    """

    def __init__(self, with_arch_param, **kwargs):
        super(DifferentiableOP, self).__init__(**kwargs)
        self.with_arch_param = with_arch_param

    def build_arch_param(self):
        """build learnable architecture parameters."""
        if self.with_arch_param:
            return nn.Parameter(torch.randn(self.num_choices) * 1e-3)
        else:
            return None

    def compute_arch_probs(self, arch_param):
        """compute chosen probs according architecture parameters."""
        return F.softmax(arch_param, -1)

    def forward(self, x, arch_param=None):
        """forward function.

        In some algorithms, there are several ``MutableModule`` share the same
        architecture parameters. So the architecture parameters are passed
        in as args.

        Args:
            prev_inputs (list[torch.Tensor]): each choice's inputs.
            arch_param (torch.nn.Parameter): architecture parameters.
        """
        if self.with_arch_param:
            assert arch_param is not None, \
                f'In {self.space_id}, the arch_param can not be None when the \
                    with_arch_param=True.'

            # 1. compute choices' probs.
            probs = self.compute_arch_probs(arch_param)

            # 2. compute every op's outputs.
            outputs = list()
            for prob, module in zip(probs, self.choice_modules):
                if prob > 0:
                    outputs.append(prob * module(x))

        else:
            outputs = list()
            for name, chosen_bool in zip(self.full_choice_names,
                                         self.choice_mask):
                if name not in self.choice_names:
                    continue
                if not chosen_bool:
                    continue
                module = self.choices[name]
                outputs.append(module(x))

            assert len(outputs) > 0
        return sum(outputs)


@MUTABLES.register_module()
class GumbelOP(DifferentiableOP):
    """Gumbel OP.

    Search the best module from choices by gumbel trick.
    """

    def __init__(self, tau=1.0, hard=True, **kwargs):
        super(GumbelOP, self).__init__(**kwargs)
        self.tau = tau
        self.hard = hard

    def set_temperature(self, temperature):
        """Modify the temperature."""
        self.temperature = temperature

    def compute_arch_probs(self, arch_param):
        """compute chosen probs by gumbel trick."""
        probs = F.gumbel_softmax(
            arch_param, tau=self.tau, hard=self.hard, dim=-1)
        return probs



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from torch.nn.modules.batchnorm import _BatchNorm

from ..builder import DISTILLERS, MODELS, build_loss
from .base import BaseDistiller


@DISTILLERS.register_module()
class SingleTeacherDistiller(BaseDistiller):
    """Distiller with single teacher.

    Args:
        teacher (dict): The config dict for teacher.
        teacher_trainable (bool): Whether the teacher is trainable.
            Default: False.
        teacher_norm_eval (bool): Whether to set teacher's norm layers to eval
            mode, namely, freeze running stats (mean and var). Note: Effect on
            Batch Norm and its variants only. Default: True.
        components (dict): The details of the distillation. It usually includes
            the module names of the teacher and the student, and the losses
            used in the distillation.
    """

    def __init__(self,
                 teacher,
                 teacher_trainable=False,
                 teacher_norm_eval=True,
                 components=tuple(),
                 **kwargs):
        super().__init__(**kwargs)
        self.teacher_trainable = teacher_trainable
        self.teacher_norm_eval = teacher_norm_eval
        self.teacher = self.build_teacher(teacher)

        self.components = components
        self.losses = nn.ModuleDict()
        self.align_modules = nn.ModuleDict()

        # Record the featuremaps that need to calculate the distillation loss.
        self.student_outputs = dict()
        self.teacher_outputs = dict()

        for i, component in enumerate(self.components):
            student_module_name = component['student_module']
            teacher_module_name = component['teacher_module']
            # The type of every student_output is a list by default, because
            # some modules will execute multiple forward calculations, such as
            # the shareable head in Retinanet
            self.student_outputs[student_module_name] = list()
            self.teacher_outputs[teacher_module_name] = list()

            # If the number of featuremap channels of student and teacher are
            # inconsistent, they need to be aligned by a 1x1 convolution
            align_module_cfg = getattr(component, 'align_module', None)
            if align_module_cfg is not None:
                align_module_name = f'component_{i}'
                align_module = self.build_align_module(align_module_cfg)
                self.align_modules[align_module_name] = align_module

            # Multiple losses can be calculated at the same location
            for loss in component.losses:
                loss_cfg = loss.copy()
                loss_name = loss_cfg.pop('name')
                self.losses[loss_name] = build_loss(loss_cfg)

    def build_teacher(self, cfg):
        """Build a model from the `cfg`."""

        teacher = MODELS.build(cfg)

        return teacher

    def build_align_module(self, cfg):
        """Build ``align_module`` from the `cfg`.

        ``align_module`` is needed when the number of channels output by the
        teacher module is not equal to that of the student module, or for some
        other reasons.

        Args:
            cfg (dict): The config dict for ``align_module``.
        """

        in_channels = cfg.student_channels
        out_channels = cfg.teacher_channels
        if cfg.type == 'conv2d':
            align_module = nn.Conv2d(in_channels, out_channels, 1)
        elif cfg.type == 'linear':
            align_module = nn.Linear(in_channels, out_channels)
        return align_module

    def prepare_from_student(self, student):
        """Registers a global forward hook for each teacher module and student
        module to be used in the distillation.

        Args:
            student (:obj:`torch.nn.Module`): The student model to be used
                in the distillation.
        """

        # Record the mapping relationship between student's modules and module
        # names.
        self.student_module2name = {}
        for name, module in student.model.named_modules():
            self.student_module2name[module] = name
        self.student_name2module = dict(student.model.named_modules())

        # Record the mapping relationship between teacher's modules and module
        # names.
        self.teacher_module2name = {}
        for name, module in self.teacher.named_modules():
            self.teacher_module2name[module] = name
        self.teacher_name2module = dict(self.teacher.named_modules())

        # Register forward hooks for modules that need to participate in loss
        # calculation.
        for component in self.components:
            student_module_name = component['student_module']
            teacher_module_name = component['teacher_module']

            student_module = self.student_name2module[student_module_name]
            teacher_module = self.teacher_name2module[teacher_module_name]

            student_module.register_forward_hook(
                self.student_forward_output_hook)
            teacher_module.register_forward_hook(
                self.teacher_forward_output_hook)

    def teacher_forward_output_hook(self, module, inputs, outputs):
        """Save the module's forward output.

        Args:
            module (:obj:`torch.nn.Module`): The module to register hook.
            inputs (tuple): The input of the module.
            outputs (tuple): The output of the module.
        """
        if self.training:
            self.teacher_outputs[self.teacher_module2name[module]].append(
                outputs)

    def student_forward_output_hook(self, module, inputs, outputs):
        """Save the module's forward output.

        Args:
            module (:obj:`torch.nn.Module`): The module to register hook.
            inputs (tuple): The input of the module.
            outputs (tuple): The output of the module.
        """
        if self.training:
            self.student_outputs[self.student_module2name[module]].append(
                outputs)

    def reset_outputs(self, outputs):
        """Reset the teacher's outputs or student's outputs."""
        for key in outputs.keys():
            outputs[key] = list()

    def exec_teacher_forward(self, data):
        """Execute the teacher's forward function.

        After this function, the teacher's featuremaps will be saved in
        ``teacher_outputs``.
        """

        # Convert the context manager's mode to teacher.
        self.reset_ctx_teacher_mode(True)
        # Clear the saved data of the last forward。
        self.reset_outputs(self.teacher_outputs)

        if self.teacher_trainable:
            output = self.teacher(**data)
        else:
            with torch.no_grad():
                output = self.teacher(**data)

        return output

    def exec_student_forward(self, student, data):
        """Execute the teacher's forward function.

        After this function, the student's featuremaps will be saved in
        ``student_outputs``.
        """
        # Convert the context manager's mode to teacher.
        self.reset_ctx_teacher_mode(False)
        # Clear the saved data of the last forward。
        self.reset_outputs(self.student_outputs)

        output = student(**data)
        return output

    def train(self, mode=True):
        """Set distiller's forward mode."""
        super(SingleTeacherDistiller, self).train(mode)
        if mode and self.teacher_norm_eval:
            for m in self.teacher.modules():
                if isinstance(m, _BatchNorm):
                    m.eval()

    def get_teacher_outputs(self, teacher_module_name):
        """Get the outputs according module name."""
        return self.teacher_outputs[teacher_module_name]

    def compute_distill_loss(self, data=None):
        """Compute the distillation loss."""

        losses = dict()

        for i, component in enumerate(self.components):
            # Get the student's outputs.
            student_module_name = component['student_module']
            student_outputs = self.student_outputs[student_module_name]

            # Align student output's channels with teacher.
            align_module_name = f'component_{i}'
            if align_module_name in self.align_modules:
                align_module = self.align_modules[align_module_name]
                student_outputs = [
                    align_module(s_out) for s_out in student_outputs
                ]

            # Get the teacher's outputs.
            teacher_module_name = component['teacher_module']
            teacher_outputs = self.get_teacher_outputs(teacher_module_name)

            # One module maybe have N outputs, such as the shareable head in
            # RetinaNet.
            for out_idx, (s_out, t_out) in enumerate(
                    zip(student_outputs, teacher_outputs)):

                for loss in component.losses:
                    loss_module = self.losses[loss.name]
                    loss_name = f'{loss.name}.{out_idx}'
                    # TODO ugly implementation.
                    # Pass the gt_label to loss function.
                    # Only used by WSLD.
                    loss_module.current_data = data
                    losses[loss_name] = loss_module(s_out, t_out)
                    loss_module.current_data = None

        return losses



# Copyright (c) OpenMMLab. All rights reserved.
import random
import warnings

import numpy as np
import torch
from mmcv.parallel import MMDataParallel, MMDistributedDataParallel
from mmcv.runner import (EpochBasedRunner, Fp16OptimizerHook, OptimizerHook,
                         build_runner)
from mmdet.core import DistEvalHook, EvalHook
from mmdet.datasets import (build_dataloader, build_dataset,
                            replace_ImageToTensor)
from mmdet.utils import get_root_logger

from mmrazor.core.distributed_wrapper import DistributedDataParallelWrapper
from mmrazor.core.hooks import DistSamplerSeedHook
from mmrazor.core.optimizer import build_optimizers
from mmrazor.utils import find_latest_checkpoint


def set_random_seed(seed, deterministic=False):
    """Import `set_random_seed` function here was deprecated in v0.3 and will
    be removed in v0.5.

    Args:
        seed (int): Seed to be used.
        deterministic (bool): Whether to set the deterministic option for
            CUDNN backend, i.e., set ``torch.backends.cudnn.deterministic``
            to True and ``torch.backends.cudnn.benchmark`` to False.
            Default: False.
    """
    warnings.warn(
        'Deprecated in v0.3 and will be removed in v0.5, '
        'please import `set_random_seed` directly from `mmrazor.apis`',
        category=DeprecationWarning)

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def train_mmdet_model(model,
                      dataset,
                      cfg,
                      distributed=False,
                      validate=False,
                      timestamp=None,
                      meta=None):
    """Copy from mmdetection and modify some codes.

    This is an ugly implementation, and will be deprecated in the future. In
    the future, there will be only one train api and no longer distinguish
    between mmclassificaiton, mmsegmentation or mmdetection.
    """

    logger = get_root_logger(cfg.log_level)

    # prepare data loaders
    dataset = dataset if isinstance(dataset, (list, tuple)) else [dataset]

    if 'imgs_per_gpu' in cfg.data:
        logger.warning('"imgs_per_gpu" is deprecated in MMDet V2.0. '
                       'Please use "samples_per_gpu" instead')
        if 'samples_per_gpu' in cfg.data:
            logger.warning(
                f'Got "imgs_per_gpu"={cfg.data.imgs_per_gpu} and '
                f'"samples_per_gpu"={cfg.data.samples_per_gpu}, "imgs_per_gpu"'
                f'={cfg.data.imgs_per_gpu} is used in this experiments')
        else:
            logger.warning(
                'Automatically set "samples_per_gpu"="imgs_per_gpu"='
                f'{cfg.data.imgs_per_gpu} in this experiments')
        cfg.data.samples_per_gpu = cfg.data.imgs_per_gpu

    runner_type = 'EpochBasedRunner' if 'runner' not in cfg else cfg.runner[
        'type']
    data_loader = [
        build_dataloader(
            ds,
            cfg.data.samples_per_gpu,
            cfg.data.workers_per_gpu,
            # cfg.gpus will be ignored if distributed
            num_gpus=len(cfg.gpu_ids),
            dist=distributed,
            seed=cfg.seed,
            runner_type=runner_type,
            persistent_workers=cfg.data.get('persistent_workers', False))
        for ds in dataset
    ]

    # put model on gpus
    if distributed:
        find_unused_parameters = cfg.get('find_unused_parameters', False)
        if cfg.get('use_ddp_wrapper', False):
            # Difference from mmdetection.
            # In some algorithms, the ``optimizer.step()`` is executed in
            # ``train_step``. To rebuilt reducer buckets rightly, there need to
            # use DistributedDataParallelWrapper.
            model = DistributedDataParallelWrapper(
                model,
                device_ids=[torch.cuda.current_device()],
                broadcast_buffers=False,
                find_unused_parameters=find_unused_parameters)
        else:
            # Sets the ``find_unused_parameters`` parameter in
            # torch.nn.parallel.DistributedDataParallel
            model = MMDistributedDataParallel(
                model.cuda(),
                device_ids=[torch.cuda.current_device()],
                broadcast_buffers=False,
                find_unused_parameters=find_unused_parameters)
    else:
        model = MMDataParallel(model, device_ids=cfg.gpu_ids)

    # build optimizers
    # Difference from mmdetection.
    # In some algorithms, there will be multi optimizers.
    optimizer = build_optimizers(model, cfg.optimizer)

    # build runner
    if 'runner' not in cfg:
        cfg.runner = {
            'type': 'EpochBasedRunner',
            'max_epochs': cfg.total_epochs
        }
        warnings.warn(
            'config is now expected to have a ``runner`` section, '
            'please set ``runner`` in your config.', UserWarning)
    else:
        if 'total_epochs' in cfg:
            assert cfg.total_epochs == cfg.runner.max_epochs

    runner = build_runner(
        cfg.runner,
        default_args=dict(
            model=model,
            optimizer=optimizer,
            work_dir=cfg.work_dir,
            logger=logger,
            meta=meta))

    # an ugly workaround to make .log and .log.json filenames the same
    runner.timestamp = timestamp

    # fp16 setting
    fp16_cfg = cfg.get('fp16', None)
    if fp16_cfg is not None:
        optimizer_config = Fp16OptimizerHook(
            **cfg.optimizer_config, **fp16_cfg, distributed=distributed)
    elif distributed and cfg.optimizer_config is not None \
            and 'type' not in cfg.optimizer_config:
        optimizer_config = OptimizerHook(**cfg.optimizer_config)
    else:
        optimizer_config = cfg.optimizer_config

    # register hooks
    runner.register_training_hooks(
        cfg.lr_config,
        optimizer_config,
        cfg.checkpoint_config,
        cfg.log_config,
        cfg.get('momentum_config', None),
        custom_hooks_config=cfg.get('custom_hooks', None))

    if distributed:
        if isinstance(runner, EpochBasedRunner):
            # Difference from mmclassification.
            # MMRazor's ``DistSamplerSeedHook`` could process multi dataloaders
            runner.register_hook(DistSamplerSeedHook())

    # register eval hooks
    if validate:
        # Support batch_size > 1 in validation
        val_samples_per_gpu = cfg.data.val.pop('samples_per_gpu', 1)
        if val_samples_per_gpu > 1:
            # Replace 'ImageToTensor' to 'DefaultFormatBundle'
            cfg.data.val.pipeline = replace_ImageToTensor(
                cfg.data.val.pipeline)
        val_dataset = build_dataset(cfg.data.val, dict(test_mode=True))
        val_dataloader = build_dataloader(
            val_dataset,
            samples_per_gpu=val_samples_per_gpu,
            workers_per_gpu=cfg.data.workers_per_gpu,
            dist=distributed,
            shuffle=False)
        eval_cfg = cfg.get('evaluation', {})
        eval_cfg['by_epoch'] = cfg.runner['type'] != 'IterBasedRunner'
        eval_hook = DistEvalHook if distributed else EvalHook
        runner.register_hook(
            eval_hook(val_dataloader, **eval_cfg), priority='LOW')

    resume_from = None
    if cfg.resume_from is None and cfg.get('auto_resume'):
        resume_from = find_latest_checkpoint(cfg.work_dir)
    if resume_from is not None:
        cfg.resume_from = resume_from

    if cfg.resume_from:
        runner.resume(cfg.resume_from)
    elif cfg.load_from:
        runner.load_checkpoint(cfg.load_from)
    runner.run(data_loader, cfg.workflow)



# Copyright (c) OpenMMLab. All rights reserved.
import copy

from mmcv.cnn import get_model_complexity_info
from mmcv.cnn.utils import revert_sync_batchnorm

from mmrazor.models.builder import ALGORITHMS
from .spos import SPOS


@ALGORITHMS.register_module()
class DetNAS(SPOS):
    """Implementation of `DetNAS <https://arxiv.org/abs/1903.10979>`_"""

    def __init__(self, **kwargs):
        super(DetNAS, self).__init__(**kwargs)

    def _init_flops(self):
        """Get flops of all modules in supernet in order to easily get each
        subnet's flops."""
        flops_model = copy.deepcopy(self.architecture)
        flops_model = revert_sync_batchnorm(flops_model)
        flops_model.eval()
        flops, params = get_model_complexity_info(flops_model.model.backbone,
                                                  self.input_shape)
        flops_lookup = dict()
        for name, module in flops_model.named_modules():
            flops = getattr(module, '__flops__', 0)
            flops_lookup[name] = flops
        del (flops_model)

        for name, module in self.architecture.named_modules():
            module.__flops__ = flops_lookup[name]



# Copyright (c) Open-MMLab. All rights reserved.

from mmcv.cnn.bricks import DropPath
from mmcv.runner import HOOKS, Hook


@HOOKS.register_module()
class DropPathProbHook(Hook):
    """Set drop_path_prob periodically.

    Args:
        max_prob (float): The max probability of dropping.
        interval (int): The saving period. If ``by_epoch=True``, interval
            indicates epochs, otherwise it indicates iterations.
            Default: -1, which means "never".
        by_epoch (bool): Saving checkpoints by epoch or by iteration.
            Default: True.
    """

    def __init__(self, max_prob, interval=-1, by_epoch=True, **kwargs):
        self.interval = interval
        self.by_epoch = by_epoch
        self.max_prob = max_prob
        assert self.by_epoch

    def before_train_epoch(self, runner):
        """Executed in before_train_epoch stage."""
        if not self.by_epoch:
            return

        if self.every_n_epochs(
                runner, self.interval) or (self.save_last
                                           and self.is_last_epoch(runner)):
            cur_epoch = runner.epoch
            max_epoch = runner._max_epochs
            drop_prob = self.max_prob * (cur_epoch * 1.0) / (max_epoch * 1.0)
            for module in runner.model.modules():
                if isinstance(module, DropPath):
                    module.drop_prob = drop_prob
            runner.logger.info(f'Set drop_prob to {drop_prob} \
                    at {runner.epoch + 1} epochs')



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from functools import partial

from mmcv.cnn import get_model_complexity_info
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.builder import ALGORITHMS
from .base import BaseAlgorithm


@ALGORITHMS.register_module()
class SPOS(BaseAlgorithm):
    """Implementation of `SPOS <https://arxiv.org/abs/1904.00420>`_"""

    def __init__(self,
                 input_shape=(3, 224, 224),
                 bn_training_mode=False,
                 **kwargs):

        super(SPOS, self).__init__(**kwargs)
        self.input_shape = input_shape
        self.bn_training_mode = bn_training_mode
        if not self.retraining:
            self._init_flops()
        self.apply(partial(self.mutator.reset_in_subnet, in_subnet=True))

    def _init_flops(self):
        """Get flops of all modules in supernet in order to easily get each
        subnet's flops."""
        flops_model = copy.deepcopy(self.architecture)
        flops_model.eval()
        if hasattr(flops_model, 'forward_dummy'):
            flops_model.forward = flops_model.forward_dummy
        else:
            raise NotImplementedError(
                'FLOPs counter is currently not currently supported with {}'.
                format(flops_model.__class__.__name__))

        flops, params = get_model_complexity_info(flops_model,
                                                  self.input_shape)
        flops_lookup = dict()
        for name, module in flops_model.named_modules():
            flops = getattr(module, '__flops__', 0)
            flops_lookup[name] = flops
        del (flops_model)

        for name, module in self.architecture.named_modules():
            module.__flops__ = flops_lookup[name]

    def get_subnet_flops(self):
        """Get subnet's flops based on the complexity information of
        supernet."""
        flops = 0
        for name, module in self.architecture.named_modules():
            if module.__in_subnet__:
                flops += getattr(module, '__flops__', 0)
        return flops

    def train_step(self, data, optimizer):
        """The iteration step during training.

        In retraining stage, to train subnet like common model. In pre-training
        stage, First to sample a subnet from supernet, then to train the
        subnet.
        """
        if self.retraining:
            outputs = super(SPOS, self).train_step(data, optimizer)
        else:
            subnet_dict = self.mutator.sample_subnet()
            self.mutator.set_subnet(subnet_dict)
            outputs = super(SPOS, self).train_step(data, optimizer)
        return outputs

    def train(self, mode=True):
        """Overwrite the train method in `nn.Module` to set `nn.BatchNorm` to
        training mode when model is set to eval mode when
        `self.bn_training_mode` is `True`.

        Args:
            mode (bool): whether to set training mode (`True`) or evaluation
                mode (`False`). Default: `True`.
        """
        super(SPOS, self).train(mode)
        if not mode and self.bn_training_mode:
            for module in self.modules():
                if isinstance(module, _BatchNorm):
                    module.training = True



# Copyright (c) OpenMMLab. All rights reserved.
from .inference import init_mmcls_model
from .train import set_random_seed, train_mmcls_model

__all__ = ['train_mmcls_model', 'init_mmcls_model', 'set_random_seed']



# Copyright (c) OpenMMLab. All rights reserved.
try:
    import mmseg
except (ImportError, ModuleNotFoundError):
    mmseg = None

if mmseg:
    from .inference import init_mmseg_model
    from .train import set_random_seed, train_mmseg_model

    __all__ = ['train_mmseg_model', 'init_mmseg_model', 'set_random_seed']



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from functools import partial

import numpy as np
import torch
import torch.distributed as dist

from mmrazor.models.builder import MUTATORS
from .base import BaseMutator


@MUTATORS.register_module()
class OneShotMutator(BaseMutator):
    """A mutator for the one-shot NAS, which mainly provide some core functions
    of changing the structure of ``ARCHITECTURES``."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def get_random_mask(space_info, searching):
        """Generate random mask for randomly sampling.

        Args:
            space_info (dict): Record the information of the space need
                to sample.
            searching (bool): Whether is in search stage.

        Returns:
            torch.Tensor: Random mask generated.
        """
        space_mask = space_info['space_mask']
        num_chosen = space_info['num_chosen']
        assert num_chosen <= space_mask.size()[0]
        choice_idx = torch.multinomial(space_mask, num_chosen)
        choice_mask = torch.zeros_like(space_mask)
        choice_mask[choice_idx] = 1
        if dist.is_available() and dist.is_initialized() and not searching:
            dist.broadcast(choice_mask, src=0)
        return choice_mask

    def sample_subnet(self, searching=False):
        """Random sample subnet by random mask.

        Args:
            searching (bool): Whether is in search stage.

        Returns:
            dict: Record the information to build the subnet from the supernet,
                its keys are the properties ``space_id`` of placeholders in the
                mutator's search spaces,
                its values are random mask generated.
        """
        subnet_dict = dict()
        for space_id, space_info in self.search_spaces.items():
            subnet_dict[space_id] = self.get_random_mask(space_info, searching)
        return subnet_dict

    def set_subnet(self, subnet_dict):
        """Setting subnet in the supernet based on the result of
        ``sample_subnet`` by changing the flag: ``in_subnet``, which is easy to
        implement some operations for subnet, such as ``forward``, calculate
        flops and so on.

        Args:
            subnet_dict (dict): Record the information to build the subnet
                from the supernet,
                its keys are the properties ``space_id`` of placeholders in the
                mutator's search spaces,
                its values are masks.
        """
        for space_id, space_info in self.search_spaces.items():
            choice_mask = subnet_dict[space_id]
            for module in space_info['modules']:
                module.choice_mask = choice_mask
                for i, choice in enumerate(module.choices.values()):
                    if choice_mask[i]:
                        choice.apply(
                            partial(self.reset_in_subnet, in_subnet=True))
                    else:
                        choice.apply(
                            partial(self.reset_in_subnet, in_subnet=False))

    @staticmethod
    def reset_in_subnet(m, in_subnet=True):
        """Reset the module's attribution.

        Args:
            m (:obj:`torch.nn.Module`): The module in the supernet.
            in_subnet (bool): If the module in subnet, set ``in_subnet`` to
                True, otherwise set to False.
        """
        m.__in_subnet__ = in_subnet

    def set_chosen_subnet(self, subnet_dict):
        """Set chosen subnet in the search_spaces after searching stage.

        Args:
            subnet_dict (dict): Record the information to build the subnet from
                the supernet,
                its keys are the properties ``space_id`` of placeholders in the
                mutator's search spaces,
                its values are masks.
        """
        for space_id, mask in subnet_dict.items():
            idxs = [i for i, x in enumerate(mask.tolist()) if x == 1.0]
            self.search_spaces[space_id]['chosen'] = [
                self.search_spaces[space_id]['choice_names'][i] for i in idxs
            ]

    def mutation(self, subnet_dict, prob=0.1):
        """Mutation used in evolution search.

        Args:
            subnet_dict (dict): Record the information to build the subnet
                from the supernet, its keys are the properties ``space_id``
                of placeholders in the mutator's search spaces, its values
                are masks.
            prob (float): The probability of mutation.

        Returns:
            dict: A new subnet_dict after mutation.
        """
        mutation_subnet_dict = copy.deepcopy(subnet_dict)
        for name, mask in subnet_dict.items():
            if np.random.random_sample() < prob:
                mutation_subnet_dict[name] = self.get_random_mask(
                    self.search_spaces[name], searching=True)
        return mutation_subnet_dict

    @staticmethod
    def crossover(subnet_dict1, subnet_dict2):
        """Crossover used in evolution search.

        Args:
            subnet_dict1 (dict): Record the information to build the subnet
                from the supernet,
                its keys are the properties ``space_id`` of placeholders in the
                mutator's search spaces,
                its values are masks.
            subnet_dict2 (dict): Record the information to build the subnet
                from the supernet,
                its keys are the properties ``space_id`` of placeholders in the
                mutator's search spaces,
                its values are masks.

        Returns:
            dict: A new subnet_dict after crossover.
        """
        crossover_subnet_dict = copy.deepcopy(subnet_dict1)
        for name, mask in subnet_dict2.items():
            if np.random.random_sample() < 0.5:
                crossover_subnet_dict[name] = mask
        return crossover_subnet_dict



# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.cnn import MODELS as MMCV_MODELS
from mmcv.utils import Registry

MODELS = Registry('models', parent=MMCV_MODELS)

ALGORITHMS = MODELS
MUTABLES = MODELS
DISTILLERS = MODELS
LOSSES = MODELS
OPS = MODELS
PRUNERS = MODELS
QUANTIZERS = MODELS
ARCHITECTURES = MODELS
MUTATORS = MODELS


def build_algorithm(cfg):
    """Build compressor."""
    return ALGORITHMS.build(cfg)


def build_architecture(cfg):
    """Build architecture."""
    return ARCHITECTURES.build(cfg)


def build_mutator(cfg):
    """Build mutator."""
    return MUTATORS.build(cfg)


def build_distiller(cfg):
    """Build distiller."""
    return DISTILLERS.build(cfg)


def build_pruner(cfg):
    """Build pruner."""
    return PRUNERS.build(cfg)


def build_mutable(cfg):
    """Build mutable."""
    return MUTABLES.build(cfg)


def build_op(cfg):
    """Build op."""
    return OPS.build(cfg)


def build_loss(cfg):
    """Build loss."""
    return LOSSES.build(cfg)



# Copyright (c) OpenMMLab. All rights reserved.
from ..builder import ARCHITECTURES
from .base import BaseArchitecture


@ARCHITECTURES.register_module()
class MMClsArchitecture(BaseArchitecture):
    """Architecture based on MMCls."""

    def __init__(self, **kwargs):
        super(MMClsArchitecture, self).__init__(**kwargs)

    def forward_dummy(self, img):
        """Used for calculating network flops."""
        output = img
        for name, child in self.model.named_children():
            if name == 'head':
                output = child.fc(output[0])
            else:
                output = child(output)
        return output

    def cal_pseudo_loss(self, pseudo_img):
        """Used for executing ``forward`` with pseudo_img."""
        return sum(pseudo_img)



# Copyright (c) OpenMMLab. All rights reserved.
from .ratio_pruning import RatioPruner
from .structure_pruning import StructurePruner
from .utils import *  # noqa: F401,F403

__all__ = ['RatioPruner', 'StructurePruner']



# Copyright (c) OpenMMLab. All rights reserved.
from abc import abstractmethod
from functools import partial

from torch import nn

from mmrazor.models.builder import MUTATORS
from mmrazor.models.mutables import MutableModule
from .base import BaseMutator


@MUTATORS.register_module()
class DifferentiableMutator(BaseMutator):
    """A mutator for the differentiable NAS, which mainly provide some core
    functions of changing the structure of ``ARCHITECTURES``."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def prepare_from_supernet(self, supernet):
        """Inherit from ``BaseMutator``'s, execute some customized functions
        exclude implementing origin ``prepare_from_supernet``.

        Args:
            supernet (:obj:`torch.nn.Module`): The architecture to be used
                in your algorithm.
        """
        super().prepare_from_supernet(supernet)
        self.arch_params = self.build_arch_params(supernet)
        self.modify_supernet_forward(supernet)

    def build_arch_params(self, supernet):
        """This function will build many arch params, which are generally used
        in diffirentiale search algorithms, such as Darts' series. Each
        space_id corresponds to an arch param, so the Mutable with the same
        space_id share the same arch param.

        Args:
            supernet (:obj:`torch.nn.Module`): The architecture to be used
                in your algorithm.

        Returns:
            torch.nn.ParameterDict: the arch params are got after traversing
                the supernet.
        """

        arch_params = nn.ParameterDict()

        # Traverse all the child modules of the model. If a child module is an
        # Space instance and its space_id is not recorded, call its
        # :func:'build_space_architecture' and record the return value. If not,
        # pass.
        def traverse(module):
            for name, child in module.named_children():

                if isinstance(child, MutableModule):
                    space_id = child.space_id
                    if space_id not in arch_params:
                        space_arch_param = child.build_arch_param()
                        if space_arch_param is not None:
                            arch_params[space_id] = space_arch_param

                traverse(child)

        traverse(supernet)

        return arch_params

    def modify_supernet_forward(self, supernet):
        """Modify the supernet's default value in forward. Traverse all child
        modules of the model, modify the supernet's default value in
        :func:'forward' of each Space.

        Args:
            supernet (:obj:`torch.nn.Module`): The architecture to be used
                in your algorithm.
        """

        def traverse(module):
            for name, child in module.named_children():

                if isinstance(child, MutableModule):
                    if child.space_id in self.arch_params.keys():
                        space_id = child.space_id
                        space_arch_param = self.arch_params[space_id]
                        child.forward = partial(
                            child.forward, arch_param=space_arch_param)

                traverse(child)

        traverse(supernet)

    @abstractmethod
    def search_subnet(self):
        pass



# Copyright (c) OpenMMLab. All rights reserved.
import copy

import mmcv
import torch
import torch.nn as nn
from mmcv.cnn import get_model_complexity_info
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.builder import ALGORITHMS, build_pruner
from mmrazor.models.utils import add_prefix
from .base import BaseAlgorithm


@ALGORITHMS.register_module()
class AutoSlim(BaseAlgorithm):
    """AutoSlim: A one-shot architecture search for channel numbers.

    Please refer to the `paper <https://arxiv.org/abs/1903.11728>` for
    details.

    Args:
        num_sample_training (int): In each iteration we train the model at
            smallest width, largest width and (`num_sample_training` − 2)
            random widths. It should be no less than 2. Defaults to 4
        input_shape (tuple): Input shape used for calculation the flops of the
            supernet.
        bn_training_mode (bool): Whether set bn to training mode when model is
            set to eval mode. Note that in slimmable networks, accumulating
            different numbers of channels results in different feature means
            and variances, which further leads to inaccurate statistics of
            shared BN. Set ``bn_training_mode`` to True to use the feature
            means and variances in a batch.
    """

    def __init__(self,
                 num_sample_training=4,
                 input_shape=(3, 224, 224),
                 bn_training_mode=False,
                 **kwargs):

        super(AutoSlim, self).__init__(**kwargs)
        assert num_sample_training >= 2, \
            'num_sample_training should be no less than 2'
        self.num_sample_training = num_sample_training

        # set bn to training mode when model is set to eval mode
        self.bn_training_mode = bn_training_mode

        if input_shape is not None:
            self.input_shape = input_shape
            self._init_flops()

    def _init_pruner(self, pruner):
        """Build registered pruners and make preparations.

        Args:
            pruner (dict): The registered pruner to be used
                in the algorithm.
        """
        if pruner is None:
            self.pruner = None
            return

        # judge whether our StructurePruner can prune the architecture
        try:
            pseudo_pruner = build_pruner(pruner)
            pseudo_architecture = copy.deepcopy(self.architecture)
            pseudo_pruner.prepare_from_supernet(pseudo_architecture)
            subnet_dict = pseudo_pruner.sample_subnet()
            pseudo_pruner.set_subnet(subnet_dict)
            subnet_dict = pseudo_pruner.export_subnet()

            pseudo_pruner.deploy_subnet(pseudo_architecture, subnet_dict)
            pseudo_img = torch.randn(1, 3, 224, 224)
            pseudo_architecture.forward_dummy(pseudo_img)
        except RuntimeError:
            raise NotImplementedError('Our current StructurePruner does not '
                                      'support pruning this architecture. '
                                      'StructurePruner is not perfect enough '
                                      'to handle all the corner cases. We will'
                                      ' appreciate it if you create a issue.')

        self.pruner = build_pruner(pruner)

        if self.retraining:
            if isinstance(self.channel_cfg, dict):
                self.pruner.deploy_subnet(self.architecture, self.channel_cfg)
                self.deployed = True
            elif isinstance(self.channel_cfg, (list, tuple)):

                self.pruner.convert_switchable_bn(self.architecture,
                                                  len(self.channel_cfg))
                self.pruner.prepare_from_supernet(self.architecture)
            else:
                raise NotImplementedError
        else:
            self.pruner.prepare_from_supernet(self.architecture)

    def _init_flops(self):
        """Get flops information of the supernet."""
        flops_model = copy.deepcopy(self.architecture)
        flops_model.eval()
        if hasattr(flops_model, 'forward_dummy'):
            flops_model.forward = flops_model.forward_dummy
        else:
            raise NotImplementedError(
                'FLOPs counter is currently not currently supported with {}'.
                format(flops_model.__class__.__name__))

        flops, params = get_model_complexity_info(
            flops_model, self.input_shape, print_per_layer_stat=False)
        flops_lookup = dict()
        for name, module in flops_model.named_modules():
            flops = getattr(module, '__flops__', 0)
            flops_lookup[name] = flops
        del (flops_model)

        for name, module in self.architecture.named_modules():
            module.__flops__ = flops_lookup[name]

    def get_subnet_flops(self):
        """A hacky way to get flops information of a subnet."""
        flops = 0
        last_out_mask_ratio = None
        for name, module in self.architecture.named_modules():
            if type(module) in [
                    nn.Conv2d, mmcv.cnn.bricks.Conv2d, nn.Linear,
                    mmcv.cnn.bricks.Linear
            ]:
                in_mask_ratio = float(module.in_mask.sum() /
                                      module.in_mask.numel())
                out_mask_ratio = float(module.out_mask.sum() /
                                       module.out_mask.numel())
                flops += module.__flops__ * in_mask_ratio * out_mask_ratio
                last_out_mask_ratio = out_mask_ratio
            elif type(module) == nn.BatchNorm2d:
                out_mask_ratio = float(module.out_mask.sum() /
                                       module.out_mask.numel())
                flops += module.__flops__ * out_mask_ratio
                last_out_mask_ratio = out_mask_ratio
            elif type(module) in [
                    nn.ReLU, nn.PReLU, nn.ELU, nn.LeakyReLU, nn.ReLU6
            ]:

                assert last_out_mask_ratio, 'An activate module can not be ' \
                                            'the first module of a network.'
                flops += module.__flops__ * last_out_mask_ratio

        return round(flops)

    def train_step(self, data, optimizer):
        """Train step function.

        This function implements the standard training iteration for
        autoslim pretraining and retraining.

        Args:
            data (dict): Input data from dataloader.
            optimizer (:obj:`torch.optim.Optimizer`): The optimizer to
                accumulate gradient
        """
        optimizer.zero_grad()

        losses = dict()
        if not self.retraining:
            assert self.pruner is not None

            self.pruner.set_max_channel()
            if self.distiller is not None:
                max_model_losses = self.distiller.exec_teacher_forward(
                    self.architecture, data)
            else:
                max_model_losses = self(**data)
            losses.update(add_prefix(max_model_losses, 'max_model'))
            max_model_loss, _ = self._parse_losses(max_model_losses)
            max_model_loss.backward()

            self.pruner.set_min_channel()
            if self.distiller is not None:
                self.distiller.exec_student_forward(self.architecture, data)
                min_model_losses = self.distiller.compute_distill_loss(data)
            else:
                min_model_losses = self(**data)
            losses.update(add_prefix(min_model_losses, 'min_model'))
            min_model_loss, _ = self._parse_losses(min_model_losses)
            min_model_loss.backward()

            for i in range(self.num_sample_training - 2):
                subnet_dict = self.pruner.sample_subnet()
                self.pruner.set_subnet(subnet_dict)
                if self.distiller is not None:
                    self.distiller.exec_student_forward(
                        self.architecture, data)
                    model_losses = self.distiller.compute_distill_loss(data)
                    losses.update(
                        add_prefix(model_losses,
                                   'prune_model{}_distiller'.format(i + 1)))
                else:
                    model_losses = self(**data)
                    losses.update(
                        add_prefix(model_losses,
                                   'prune_model{}'.format(i + 1)))
                model_loss, _ = self._parse_losses(model_losses)
                model_loss.backward()
        else:
            if self.deployed:
                # Only one subnet retrains. The supernet has already deploy
                model_losses = self(**data)
                losses.update(add_prefix(model_losses, 'prune_model'))
                model_loss, _ = self._parse_losses(model_losses)
                model_loss.backward()
            else:
                # More than one subnet retraining together
                assert isinstance(self.channel_cfg, (list, tuple))
                for i, subnet in enumerate(self.channel_cfg):
                    self.pruner.switch_subnet(subnet, i)
                    model_losses = self(**data)
                    losses.update(
                        add_prefix(model_losses,
                                   'prune_model_{}'.format(i + 1)))
                    model_loss, _ = self._parse_losses(model_losses)
                    model_loss.backward()

        # TODO: clip grad norm
        optimizer.step()

        loss, log_vars = self._parse_losses(losses)
        outputs = dict(
            loss=loss, log_vars=log_vars, num_samples=len(data['img'].data))

        return outputs

    def train(self, mode=True):
        """Overwrite the train method in ``nn.Module`` to set ``nn.BatchNorm``
        to training mode when model is set to eval mode when
        ``self.bn_training_mode`` is ``True``.

        Args:
            mode (bool): whether to set training mode (``True``) or evaluation
                mode (``False``). Default: ``True``.
        """
        super(AutoSlim, self).train(mode)
        if not mode and self.bn_training_mode:
            for module in self.modules():
                if isinstance(module, _BatchNorm):
                    module.training = True
